# This program handles the communication over I2C
# between a Raspberry Pi and a combo.
# Made by: MrTijn/Tijndagamer
# Released under the MIT License
# Copyright (c) 2015, 2016, 2017 MrTijn/Tijndagamer
from robophery.interface.i2c import I2cModule


class Mpu6050Module(I2cModule):
    """
    MPU6050 PWM MPU-6050 Gyroscope / Accelerometer sensor.
    """
    DEVICE_NAME = 'mpu6050'
    DEVICE_ADDR = 0x68

    # Global Variables
    GRAVITIY_MS2 = 9.80665

    # Scale Modifiers
    ACCEL_SCALE_MODIFIER_2G = 16384.0
    ACCEL_SCALE_MODIFIER_4G = 8192.0
    ACCEL_SCALE_MODIFIER_8G = 4096.0
    ACCEL_SCALE_MODIFIER_16G = 2048.0

    GYRO_SCALE_MODIFIER_250DEG = 131.0
    GYRO_SCALE_MODIFIER_500DEG = 65.5
    GYRO_SCALE_MODIFIER_1000DEG = 32.8
    GYRO_SCALE_MODIFIER_2000DEG = 16.4

    # Pre-defined ranges
    ACCEL_RANGE_2G = 0x00
    ACCEL_RANGE_4G = 0x08
    ACCEL_RANGE_8G = 0x10
    ACCEL_RANGE_16G = 0x18

    GYRO_RANGE_250DEG = 0x00
    GYRO_RANGE_500DEG = 0x08
    GYRO_RANGE_1000DEG = 0x10
    GYRO_RANGE_2000DEG = 0x18

    # MPU-6050 Registers
    PWR_MGMT_1 = 0x6B
    PWR_MGMT_2 = 0x6C

    ACCEL_XOUT0 = 0x3B
    ACCEL_YOUT0 = 0x3D
    ACCEL_ZOUT0 = 0x3F

    TEMP_OUT0 = 0x41

    GYRO_XOUT0 = 0x43
    GYRO_YOUT0 = 0x45
    GYRO_ZOUT0 = 0x47

    ACCEL_CONFIG = 0x1C
    GYRO_CONFIG = 0x1B

    def __init__(self, *args, **kwargs):
        super(Mpu6050Module, self).__init__(*args, **kwargs)
        self._data = self._setup_i2c_iface(kwargs.get('data'))
        # Wake up the MPU-6050 since it starts in sleep mode
        self._data.write8(self.PWR_MGMT_1, 0x00)

    def read_word(self, register):
        """
        Read two I2C registers and combine them.
        Returns the combined read results.
        register - the first register to read from.
        """
        # Read the data from the registers
        high = self._data.readU8(register)
        low = self._data.readU8(register + 1)

        value = (high << 8) + low

        if (value >= 0x8000):
            return -((65535 - value) + 1)
        else:
            return value

    # MPU-6050 Methods

    def get_temp(self):
        """
        Reads the temperature from the onboard temperature sensor
        of the MPU-6050. Returns the temperature in degrees Celcius.
        """
        raw_temp = self.read_word(self.TEMP_OUT0)

        # Get the actual temperature using the formule given in the
        # MPU-6050 Register Map and Descriptions revision 4.2, page 30
        actual_temp = (raw_temp / 340.0) + 36.53

        return actual_temp

    def set_accel_range(self, accel_range):
        """
        Sets the range of the accelerometer to range.
        accel_range -- the range to set the accelerometer to. Using a
        pre-defined range is advised.
        """
        # First change it to 0x00 to make sure we write the correct value later
        self._data.write8(self.ACCEL_CONFIG, 0x00)
        # Write the new range to the ACCEL_CONFIG register
        self._data.write8(self.ACCEL_CONFIG, accel_range)

    def read_accel_range(self, raw=False):
        """Reads the range the accelerometer is set to.
        If raw is True, it will return the raw value from the ACCEL_CONFIG
        register
        If raw is False, it will return an integer: -1, 2, 4, 8 or 16. When it
        returns -1 something went wrong.
        """
        raw_data = self._data.readU8(self.ACCEL_CONFIG)

        if raw is True:
            return raw_data
        elif raw is False:
            if raw_data == self.ACCEL_RANGE_2G:
                return 2
            elif raw_data == self.ACCEL_RANGE_4G:
                return 4
            elif raw_data == self.ACCEL_RANGE_8G:
                return 8
            elif raw_data == self.ACCEL_RANGE_16G:
                return 16
            else:
                return -1

    def get_accel_data(self, g=False):
        """Gets and returns the X, Y and Z values from the accelerometer.
        If g is True, it will return the data in g
        If g is False, it will return the data in m/s^2
        Returns a dictionary with the measurement results.
        """
        x = self.read_word(self.ACCEL_XOUT0)
        y = self.read_word(self.ACCEL_YOUT0)
        z = self.read_word(self.ACCEL_ZOUT0)

        accel_scale_modifier = None
        accel_range = self.read_accel_range(True)

        if accel_range == self.ACCEL_RANGE_2G:
            accel_scale_modifier = self.ACCEL_SCALE_MODIFIER_2G
        elif accel_range == self.ACCEL_RANGE_4G:
            accel_scale_modifier = self.ACCEL_SCALE_MODIFIER_4G
        elif accel_range == self.ACCEL_RANGE_8G:
            accel_scale_modifier = self.ACCEL_SCALE_MODIFIER_8G
        elif accel_range == self.ACCEL_RANGE_16G:
            accel_scale_modifier = self.ACCEL_SCALE_MODIFIER_16G
        else:
            self._log.error("Unkown range for accel_scale_modifier.")
            accel_scale_modifier = self.ACCEL_SCALE_MODIFIER_2G

        x = x / accel_scale_modifier
        y = y / accel_scale_modifier
        z = z / accel_scale_modifier

        if g is True:
            return {'x': x, 'y': y, 'z': z}
        elif g is False:
            x = x * self.GRAVITIY_MS2
            y = y * self.GRAVITIY_MS2
            z = z * self.GRAVITIY_MS2
            return {'x': x, 'y': y, 'z': z}

    def set_gyro_range(self, gyro_range):
        """
        Sets the range of the gyroscope to range.
        gyro_range -- the range to set the gyroscope to. Using a pre-defined
        range is advised.
        """
        # First change it to 0x00 to make sure we write the correct value later
        self._data.write8(self.GYRO_CONFIG, 0x00)

        # Write the new range to the ACCEL_CONFIG register
        self._data.write8(self.GYRO_CONFIG, gyro_range)

    def read_gyro_range(self, raw=False):
        """
        Reads the range the gyroscope is set to.
        If raw is True, it will return the raw value from the GYRO_CONFIG
        register.
        If raw is False, it will return 250, 500, 1000, 2000 or -1. If the
        returned value is equal to -1 something went wrong.
        """
        raw_data = self._data.readU8(self.GYRO_CONFIG)

        if raw is True:
            return raw_data
        elif raw is False:
            if raw_data == self.GYRO_RANGE_250DEG:
                return 250
            elif raw_data == self.GYRO_RANGE_500DEG:
                return 500
            elif raw_data == self.GYRO_RANGE_1000DEG:
                return 1000
            elif raw_data == self.GYRO_RANGE_2000DEG:
                return 2000
            else:
                return -1

    def get_gyro_data(self):
        """
        Gets and returns the X, Y and Z values from the gyroscope.
        Returns the read values in a dictionary.
        """
        x = self.read_word(self.GYRO_XOUT0)
        y = self.read_word(self.GYRO_YOUT0)
        z = self.read_word(self.GYRO_ZOUT0)

        gyro_scale_modifier = None
        gyro_range = self.read_gyro_range(True)

        if gyro_range == self.GYRO_RANGE_250DEG:
            gyro_scale_modifier = self.GYRO_SCALE_MODIFIER_250DEG
        elif gyro_range == self.GYRO_RANGE_500DEG:
            gyro_scale_modifier = self.GYRO_SCALE_MODIFIER_500DEG
        elif gyro_range == self.GYRO_RANGE_1000DEG:
            gyro_scale_modifier = self.GYRO_SCALE_MODIFIER_1000DEG
        elif gyro_range == self.GYRO_RANGE_2000DEG:
            gyro_scale_modifier = self.GYRO_SCALE_MODIFIER_2000DEG
        else:
            self._log.error("Unkown range for gyro_scale_modifier.")
            gyro_scale_modifier = self.GYRO_SCALE_MODIFIER_250DEG

        x = x / gyro_scale_modifier
        y = y / gyro_scale_modifier
        z = z / gyro_scale_modifier

        return {'x': x, 'y': y, 'z': z}

    def read_data(self):
        """
        Get all sensor readings.
        """
        temp_time_start = self._get_time()
        temp = self.get_temp()
        temp_time_stop = self._get_time()
        temp_time_delta = temp_time_stop - temp_time_start
        accel_time_start = self._get_time()
        accel = self.get_accel_data()
        accel_time_stop = self._get_time()
        accel_time_delta = (accel_time_stop - accel_time_start) / 3
        gyro_time_start = self._get_time()
        gyro = self.get_gyro_data()
        gyro_time_stop = self._get_time()
        gyro_time_delta = (gyro_time_stop - gyro_time_start) / 3
        data = [
            (self._name, 'temperature', temp, temp_time_delta),
            (self._name, 'rotation_x', gyro['x'], gyro_time_delta),
            (self._name, 'rotation_y', gyro['y'], gyro_time_delta),
            (self._name, 'rotation_z', gyro['z'], gyro_time_delta),
            (self._name, 'acceleration_x', accel['x'], accel_time_delta),
            (self._name, 'acceleration_y', accel['y'], accel_time_delta),
            (self._name, 'acceleration_z', accel['z'], accel_time_delta),
        ]
        self._log_data(data)
        return data

    def meta_data(self):
        """
        Get the readings meta-data.
        """
        return {
            'temperature': {
                'type': 'gauge',
                'unit': 'C',
                'precision': 0.4,
                'range_low': -10,
                'range_high': 85,
                'sensor': self.DEVICE_NAME
            },
            'rotation_x': {
                'type': 'gauge',
                'unit': 'deg',
                'precision': 1,
                'range_low': 0,
                'range_high': 180,
                'sensor': self.DEVICE_NAME
            },
            'rotation_y': {
                'type': 'gauge',
                'unit': 'deg',
                'precision': 1,
                'range_low': 0,
                'range_high': 180,
                'sensor': self.DEVICE_NAME
            },
            'rotation_z': {
                'type': 'gauge',
                'unit': 'deg',
                'precision': 1,
                'range_low': 0,
                'range_high': 180,
                'sensor': self.DEVICE_NAME
            },
            'acceleration_x': {
                'type': 'gauge',
                'unit': 'g',
                'precision': 0.01,
                'range_low': 0,
                'range_high': 20,
                'sensor': self.DEVICE_NAME
            },
            'acceleration_y': {
                'type': 'gauge',
                'unit': 'g',
                'precision': 0.01,
                'range_low': 0,
                'range_high': 20,
                'sensor': self.DEVICE_NAME
            },
            'acceleration_z': {
                'type': 'gauge',
                'unit': 'g',
                'precision': 0.01,
                'range_low': 0,
                'range_high': 20,
                'sensor': self.DEVICE_NAME
            },
        }
